/**
 * Copyright 2004-present Facebook. All Rights Reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @flow
 * @format
 */
'use strict';

const autoprefixer = require('autoprefixer');
const paths = require('./paths');
const webpack = require('webpack');
const BundleAnalyzerPlugin = require('webpack-bundle-analyzer')
  .BundleAnalyzerPlugin;
const ManifestPlugin = require('webpack-manifest-plugin');
const TerserPlugin = require('terser-webpack-plugin');

type Options = {
  projectName?: string,
  extraPaths?: string[],
  entry?: Object,
  devtool?: string,
};

function createProductionWebpackConfig(options: Options) {
  return {
    mode: 'production',
    devtool: options.devtool ?? false,
    entry: Object.assign(
      {
        main: [paths.appIndexJs],
        login: [paths.loginJs],
      },
      options.entry || {},
    ),
    externals: [
      {
        xmlhttprequest: '{XMLHttpRequest:XMLHttpRequest}',
      },
    ],
    module: {
      rules: [
        {
          // "oneOf" will traverse all following loaders until one will
          // match the requirements. When no loader matches it will fall
          // back to the "file" loader at the end of the loader list.
          oneOf: [
            // "url" loader works like "file" loader except that it embeds
            // assets smaller than specified limit in bytes as data URLs to
            // avoid requests.  A missing `test` is equivalent to a match.
            {
              loader: require.resolve('url-loader'),
              options: {
                limit: 10000,
                name: 'static/media/[name].[hash:8].[ext]',
              },
              test: [/\.bmp$/, /\.gif$/, /\.jpe?g$/, /\.png$/],
            },
            // Process JS with Babel.
            {
              include: [
                paths.appSrc,
                paths.packagesDir,
                ...(options.extraPaths || []),
              ],
              loader: require.resolve('babel-loader'),
              options: {
                configFile: '../../babel.config.js',
                // This is a feature of `babel-loader` for webpack (not Babel
                // itself). It enables caching results in
                // ./node_modules/.cache/babel-loader/ directory for faster
                // rebuilds.
                cacheDirectory: true,
              },
              test: /\.(js|jsx|mjs)$/,
            },
            // "postcss" loader applies autoprefixer to our CSS.
            // "css" loader resolves paths in CSS and adds assets as
            // dependencies.
            // "style" loader turns CSS into JS modules that inject <style>
            // tags.
            // In production, we use a plugin to extract that CSS to a file, but
            // in development "style" loader enables hot editing of CSS.
            {
              test: /\.css$/,
              use: [
                require.resolve('style-loader'),
                {
                  loader: require.resolve('css-loader'),
                  options: {
                    importLoaders: 1,
                  },
                },
                {
                  loader: require.resolve('postcss-loader'),
                  options: {
                    // Necessary for external CSS imports to work
                    // https://github.com/facebookincubator/create-react-app/issues/2677
                    ident: 'postcss',
                    plugins: () => [
                      require('postcss-flexbugs-fixes'),
                      autoprefixer({
                        browsers: [
                          '>1%',
                          'last 4 versions',
                          'Firefox ESR',
                          'not ie < 9', // React doesn't support IE8 anyway
                        ],
                        flexbox: 'no-2009',
                      }),
                    ],
                  },
                },
              ],
            },
            // "file" loader makes sure those assets get served by
            // WebpackDevServer.
            // When you `import` an asset, you get its (virtual) filename.
            // In production, they would get copied to the `build` folder.
            // This loader doesn't use a "test" so it will catch all modules
            // that fall through the other loaders.
            {
              // Exclude `js` files to keep "css" loader working as it injects
              // its runtime that would otherwise processed through "file"
              // loader. Also exclude `html` and `json` extensions so they get
              // processed by webpacks internal loaders.
              exclude: [/\.(js|jsx|mjs)$/, /\.html$/, /\.json$/],
              loader: require.resolve('file-loader'),
              options: {
                name: 'static/media/[name].[hash:8].[ext]',
              },
            },
          ],
        },
        // ** STOP ** Are you adding a new loader?
        // Make sure to add the new loader(s) before the "file" loader.
      ],
    },
    // Some libraries import Node modules but don't use them in the browser.
    // Tell Webpack to provide empty mocks for them so importing them works.
    node: {
      child_process: 'empty',
      dgram: 'empty',
      fs: 'empty',
      net: 'empty',
      tls: 'empty',
    },
    output: {
      chunkFilename: 'static/js/[name].chunk.js',
      filename: '[name].[chunkhash].js',
      path: paths.distPath,
      pathinfo: true,
      publicPath: options.projectName
        ? `/${options.projectName}/static/dist/`
        : '/static/dist/',
    },
    plugins: [
      new ManifestPlugin(),
      new BundleAnalyzerPlugin({
        analyzerMode: 'static',
        reportFilename: 'report.html',
      }),
      new webpack.NoEmitOnErrorsPlugin(),
      // remove excess locales in moment bloating the bundle
      new webpack.ContextReplacementPlugin(/moment[/\\]locale$/, /en/),
    ],
    optimization: {
      minimizer: [
        new TerserPlugin({
          chunkFilter: chunk => {
            if (chunk.name === 'vendor') {
              return false;
            }
            return true;
          },
          parallel: true,
        }),
      ],
      splitChunks: {
        cacheGroups: {
          vendor: {
            test: /[\\/]node_modules[\\/]/,
            chunks: 'all',
            name: 'vendor',
            filename: 'vendor.[chunkhash].js',
          },
        },
      },
    },
  };
}

module.exports = {
  createProductionWebpackConfig,
};
